from setuptools import setup

setup(
    install_requires=[
        "numpy>=1.21.1",
        "numfi>=0.2.4",
        "matplotlib>=3.5.2",
        "pandas>=1.3.1",
        "openpyxl>=3.0.7",
        "h5py>=3.7.0",
        "PyYAML>=5.4.1",
        "scipy>=1.6.3"
    ]
)
